@@ -48,15 +48,15 @@ PLATFORMS_DEVICES = cli_args.platforms_devices
 #
 # devices
 #
-devices: list[(int, int, Any, Any)] = []
+devices: list[tuple[int, int, int]] = []
 devices_locks: list[asyncio.Lock] = []
-# devices_procs: list[tuple[str | None, asyncio.subprocess.Process | None]] = []
+devices_wss: dict[tuple[int, int, int], Any] = {}
+devices_procs: dict[tuple[int, int, int], asyncio.subprocess.Process] = {}
 task_queue = set()
 
 def init_devices():
     global devices
     global devices_locks
-    # global devices_procs
     
     if BACKEND == 'cpu':
         # allow only ONE "device" concurrently
@@ -79,10 +79,8 @@ def init_devices():
     for n in devices:
         devices_locks.append((n, asyncio.Lock()))
 
-    # create devices_procs and set all values to None
-    # devices_procs = [(None, None)] * len(devices)
-    print('devices_locks:', devices_locks)
-    # print('devices_procs:', devices_procs)
+    print('devices_locks:')
+    pprint(devices_locks)
 
 def parse_llama_print_timings(text: str) -> dict:
     lines = [n for n in text.splitlines() if n.startswith('llama_print_timings:')]
@@ -107,8 +105,6 @@ def parse_llama_print_timings(text: str) -> dict:
             for n in v
         }
 
-    # print('parse_llama_print_timings:')
-    # print('lines:', lines)
     return lines
 
 def build_llama_cpp_cmd(device: tuple[int, int, int],
@@ -150,7 +146,7 @@ def build_llama_cpp_cmd(device: tuple[int, int, int],
         '--top-k', top_k,
         '--top-p', top_p,
         # '--mlock',
-        # '--no-mmap',
+        '--no-mmap',
         '--simple-io',
         '--log-disable',
         '--prompt', shell_prompt,
@@ -185,8 +181,6 @@ async def run_prompt(device: tuple[int, int, int],
     proc_model: str | None = None
     proc: asyncio.subprocess.Process | None = None
 
-    # print('? prompt:', repr(prompt))
-
     cmd: str = build_llama_cpp_cmd(
         device=device,
         prompt=prompt,
@@ -202,25 +196,6 @@ async def run_prompt(device: tuple[int, int, int],
 
     try:
         async with timeout(TIMEOUT) as cm:
-            # get eager proc with its last used model
-            # proc_model, proc = devices_procs[index]
-            #
-            # if proc_model != model:
-            #     if proc:
-            #         # close proc because model is wrong
-            #         proc.kill()
-            #         await proc.wait()
-            #         print('proc kill [wrong model]')
-            #
-            #     # create new proc for model
-            #     proc = await asyncio.create_subprocess_shell(
-            #         cmd,
-            #         stdout=asyncio.subprocess.PIPE,
-            #         stderr=asyncio.subprocess.PIPE,
-            #     )
-            #
-            #     devices_procs[index] = (model, proc)
-            #     print('devices_procs:', devices_procs)
             # create new proc for model
             t0: float = time.time()
 
@@ -230,6 +205,9 @@ async def run_prompt(device: tuple[int, int, int],
                 stderr=asyncio.subprocess.PIPE,
             )
             
+            # map device to proc
+            devices_procs[device] = proc
+
             stdout: bytes
             stderr: bytes
 
@@ -249,7 +227,7 @@ async def run_prompt(device: tuple[int, int, int],
                     if len(stdout) > len(prompt_enc):
                         break
 
-                    await asyncio.sleep(0.1)
+                    await asyncio.sleep(0.2)
 
                 stdout = stdout[1 + len(prompt_enc):]
                 stderr = b''
@@ -261,6 +239,7 @@ async def run_prompt(device: tuple[int, int, int],
                 buf = stdout
                 prev_buf = b''
                 text = stdout.decode()
+                print(f'[{index}, {pi}, {di}]:', repr(text))
 
                 res = {
                     'id': id_,
@@ -284,7 +263,8 @@ async def run_prompt(device: tuple[int, int, int],
                         continue
 
                     prev_buf = b''
-                    
+                    print(f'[{index}, {pi}, {di}]:', repr(text))
+
                     res = {
                         'id': id_,
                         'status': 'chunk',
@@ -306,15 +286,18 @@ async def run_prompt(device: tuple[int, int, int],
                     if stopped:
                         break
 
-                    await asyncio.sleep(0.1)
+                    await asyncio.sleep(0.2)
             else:
                 stdout, stderr = await proc.communicate()
                 stdout = stdout[1 + len(prompt_enc):]
 
             if streaming:
-                stderr = await proc.stderr.read()
+                if not stopped:
+                    stderr = await proc.stderr.read()
 
             if stopped:
+                print('stopword, trying to kill proc', proc.pid)
+
                 try:
                     proc.kill()
                     await proc.wait()
@@ -327,6 +310,8 @@ async def run_prompt(device: tuple[int, int, int],
             stdout = stdout.decode().strip()
             stderr = stderr.decode().strip()
     except asyncio.TimeoutError as e:
+        print('timeout, trying to kill proc', proc.pid)
+
         try:
             proc.kill()
             await proc.wait()
@@ -336,24 +321,11 @@ async def run_prompt(device: tuple[int, int, int],
         finally:
             proc = None
 
-    # print('!! stdout', repr(stdout))
-    # print('!! stderr', repr(stderr))
-
     print('!! stdout:')
     print(stdout)
     
-    print('!! stderr:')
-    print(stderr)
-
-    # create eager proc for model
-    # proc = await asyncio.create_subprocess_shell(
-    #     cmd,
-    #     stdout=asyncio.subprocess.PIPE,
-    #     stderr=asyncio.subprocess.PIPE,
-    # )
-    #
-    # devices_procs[index] = (model, proc)
-    # print('devices_procs:', devices_procs)
+    # print('!! stderr:')
+    # print(stderr)
 
     if cm.expired:
         res = {
@@ -367,6 +339,10 @@ async def run_prompt(device: tuple[int, int, int],
     else:
         yield True, stdout, stderr
 
+    # remove map device to proc
+    if device in devices_procs:
+        del devices_procs[device]
+
 #
 # web server
 #
@@ -503,6 +479,9 @@ async def _ws_text_completion_stream(ws, id_: str, data: dict):
         if lock:
             break
 
+    # map device to ws
+    devices_wss[device] = ws
+
     # run prompt
     stdout = None
     stderr = None
@@ -579,30 +558,55 @@ async def _ws_text_completion_stream(ws, id_: str, data: dict):
     await ws.send_json(res)
     await ws.close()
 
+    # remove map device to ws
+    if device in devices_wss:
+        del devices_wss[device]
+
 @routes.get('/api/1.0/text/completion')
 async def get_api_1_0_text_completion(request, id_: str):
     ws = web.WebSocketResponse()
     await ws.prepare(request)
     print('websocket connection opened')
 
-    async with asyncio.TaskGroup() as tg:
-        async for msg in ws:
-            if msg.type == aiohttp.WSMsgType.PING:
-                await ws.pong(msg.data)
-            elif msg.type == aiohttp.WSMsgType.TEXT:
-                data = json.loads(msg.data)
+    try:
+        async with asyncio.TaskGroup() as tg:
+            async for msg in ws:
+                if msg.type == aiohttp.WSMsgType.PING:
+                    await ws.pong(msg.data)
+                elif msg.type == aiohttp.WSMsgType.TEXT:
+                    data = json.loads(msg.data)
 
-                # run in background
-                coro = _ws_text_completion_stream(ws, id_, data)
-                task = tg.create_task(coro)
-            elif msg.type == aiohttp.WSMsgType.ERROR:
-                print(f'ws connection closed with exception {ws.exception()}')
-                await ws.close()
-                break
-            else:
-                print(f'ws msg.type:', msg.type)
+                    # run in background
+                    coro = _ws_text_completion_stream(ws, id_, data)
+                    task = tg.create_task(coro)
+                elif msg.type == aiohttp.WSMsgType.ERROR:
+                    print(f'ws connection closed with exception {ws.exception()}')
+                    await ws.close()
+                    break
+                else:
+                    print(f'ws msg.type:', msg.type)
+
+            await task
+    except ExceptionGroup as e:
+        print('ExceptionGroup:', e)
+
+        # find device
+        device = [(k, v) for k, v in devices_wss.items() if v is ws][0][0]
+        print('kill proc for device:', device)
+
+        # kill proc
+        while device not in devices_procs:
+            await asyncio.sleep(0.5)
+
+        proc = devices_procs[device]
+        proc.kill()
+        await proc.wait()
+        del devices_procs[device]
+
+        # close ws
+        await ws.close()
+        del devices_wss[device]
 
-    await task
     print('websocket connection closed')
     return ws
 
